/**
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 2009-2012 Bull S.A.S.
 * Contact: jonas-team@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

package org.ow2.jonas.admin.jdbcconnections.service;

import java.io.IOException;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Set;

import javax.management.Attribute;
import javax.management.ObjectName;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.apache.felix.ipojo.annotations.Component;
import org.apache.felix.ipojo.annotations.Instantiate;
import org.apache.felix.ipojo.annotations.Provides;
import org.apache.felix.ipojo.annotations.ServiceProperty;
import org.apache.felix.ipojo.annotations.Validate;
import org.ow2.kerneos.core.service.KerneosSimpleService;
import org.ow2.util.log.Log;
import org.ow2.util.log.LogFactory;
import org.w3c.dom.Document;
import org.w3c.dom.Element;


@Component
@Instantiate
@Provides
public class JDBCConnectionsService extends JMXHelperService implements KerneosSimpleService, Serializable {

    @ServiceProperty(name = KerneosSimpleService.ID, value = "jdbcConnectionsService")
    private String id;

    /**
     * Logger of the class
     */
    private Log logger = LogFactory.getLog(JDBCConnectionsService.class);


    /**
     * JNDI Interceptor ObjectName.
     */
    private ObjectName jdbcJNDIInterceptorObjectName = null;

    /**
     * Default constructor.
     */
    public JDBCConnectionsService() {
        super();
    }

    @Validate
    private void start() {
        logger.debug("");
        this.jdbcJNDIInterceptorObjectName = buildObjectName(getDomainName() + ":type=service,name=jndi-interceptors,interceptor=org.ow2.jonas.jndi.interceptors.impl.datasource.DataSourceLeakDetectorContextInterceptor");
    }

    /**
     * @return the information about the JDBC JNDI Interceptor (like regexp pattern, etc)
     */
    public JNDIInterceptorsConfiguration getJNDIInterceptorsConfiguration() {
        // Component is enabled if MBeans is registered ?
        JNDIInterceptorsConfiguration jndiInterceptorsConfiguration = new JNDIInterceptorsConfiguration();
       if (isRegistered()) {
            jndiInterceptorsConfiguration.setJndiInterceptorEnabled(true);

            // Component is available, look at the configuration
            // Force closing connections ?
            Boolean forceClose = getAttribute(jdbcJNDIInterceptorObjectName, "forceClose");
            jndiInterceptorsConfiguration.setForceClose(forceClose.booleanValue());

            // Regexp pattern
            String regExpPattern = getAttribute(jdbcJNDIInterceptorObjectName, "jndiRegexp");
            jndiInterceptorsConfiguration.setRegExp(regExpPattern);
        }
        return jndiInterceptorsConfiguration;
    }

    /**
     * Update the regexp pattern of the JDBC JNDI Interceptor
     * @param newRegExpValue the new value
     */
    public void updateRegExp(final String newRegExpValue) {
        if (!isRegistered()) {
            // Cannot update
            throw new IllegalStateException("Component not available, attribute cannot be updated");
        }

        //  Update value
        setAttribute(jdbcJNDIInterceptorObjectName, new Attribute("jndiRegexp", newRegExpValue));
    }

    /**
     * Update the forceClose option of the JDBC JNDI Interceptor
     * @param forceClose true if needs to close the connection
     */
    public void updateForceClose(final boolean forceClose) {
        if (!isRegistered()) {
            // Cannot update
            throw new IllegalStateException("Component not available, attribute cannot be updated");
        }

        //  Update value
        setAttribute(jdbcJNDIInterceptorObjectName, new Attribute("forceClose", forceClose));
    }

    /**
     * Update the observable attribute on the given datasource.
     * @param objectNameString the datasource JMX objectname
     * @param observable true/false
     */
    public void updateObservable(final String objectNameString, final boolean observable) {

        // Get Object
        ObjectName objectName = buildObjectName(objectNameString);
        if (!isRegistered(objectName)) {
            throw new IllegalStateException("No such ObjectName with value '" + objectNameString + "'.");
        }

        if (!isRegistered(objectName)) {
            // Cannot update
            throw new IllegalStateException("Component not available, attribute cannot be updated");
        }

        //  Update value
        setAttribute(objectName, new Attribute("observable", observable));
    }


    /**
     * @return an XML document with all dbm/resource service datasources.
     */
    public Document getAvailableJDBCDatasources() {

            // Create builder with factory
            DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
            DocumentBuilder builder = null;
            try {
                builder = factory.newDocumentBuilder();
            } catch (ParserConfigurationException e) {
                throw new IllegalStateException("Cannot build document builder", e);
            }

            // Create Document
            Document document = builder.newDocument();

            // Append root element
            Element root = document.createElement("root");
            document.appendChild(root);

            // Folder name
            Element nodeElement = document.createElement("node");
            nodeElement.setAttribute("label", "JDBC Datasources");
            nodeElement.setAttribute("isBranch", "true");
            root.appendChild(nodeElement);

            // Get All JDBC datasources
            Set<ObjectName> datasourceMBeans = new HashSet<ObjectName>();

            // Get DBM Datasources
            ObjectName dbmDatasourceNames = buildObjectName(getDomainName() + ":j2eeType=JDBCDataSource,*");
            Set<ObjectName> dbmDatasourceMBeans = null;
            try {
                dbmDatasourceMBeans = getMBeanServerConnection().queryNames(dbmDatasourceNames, null);
            } catch (IOException e) {
                throw new IllegalStateException("Cannot get ObjectName of dbm datasources", e);
            }
            datasourceMBeans.addAll(dbmDatasourceMBeans);


            // Add the Resource services datasources
            ObjectName resourcesdatasourceNames = buildObjectName(getDomainName() + ":j2eeType=JCAConnectionFactory,*");
            Set<ObjectName> resourceDatasourceMBeans = null;
            try {
                resourceDatasourceMBeans = getMBeanServerConnection().queryNames(resourcesdatasourceNames, null);
            } catch (IOException e) {
                throw new IllegalStateException("Cannot get ObjectName of resource service datasources", e);
            }
            // check if it is a datasource
            Iterator<ObjectName> itResourceMBean = resourceDatasourceMBeans.iterator();
            while (itResourceMBean.hasNext()) {
                ObjectName resourceDatasourceMBean = itResourceMBean.next();
                Properties p = getAttribute(resourceDatasourceMBean, "properties");
                //This is a Datasource Resource factory !
                if (p.getProperty("dsClass") != null && !"".equals(p.getProperty("dsClass"))) {
                    datasourceMBeans.add(resourceDatasourceMBean);
                }
            }

            // Iterator
            Iterator<ObjectName> it = datasourceMBeans.iterator();

            // Loop on all ObjectName
            while (it.hasNext()) {
                ObjectName objectName = it.next();

                // Get JNDI Name
                String jndiName = getAttribute(objectName, "jndiName");

                // Create elements
                Element itemElement = document.createElement("item");
                itemElement.setAttribute("label", jndiName);
                itemElement.setAttribute("objectName", objectName.toString());
                nodeElement.appendChild(itemElement);
            }

            return document;
    }

    /**
     * @param objectNameString the ObjectName of the datasource.
     * @return datasource information on the given Datasource
     */
    public JDBCDatasourceInfo getJDBCDatasourceInfo(final String objectNameString) {

        // Get Object
        ObjectName objectName = buildObjectName(objectNameString);
        if (!isRegistered(objectName)) {
            throw new IllegalStateException("No such ObjectName with value '" + objectNameString + "'.");
        }

        // DBM datasource ?
        if (objectNameString.indexOf("JDBCDataSource") != -1) {
            return getDBMDatasourceInfo(objectName);
        } else if (objectNameString.indexOf("JCAConnectionFactory") != -1) {
            return getResourceServiceDatasourceInfo(objectName);
        }
        throw new IllegalArgumentException("Invalid ObjectName '" + objectNameString + "'.");
    }

    /**
     * @param objectName the JMX objectname of the datasource.
     * @return the datasource information on this datasouce
     */
    public JDBCDatasourceInfo getDBMDatasourceInfo(final ObjectName objectName) {
        JDBCDatasourceInfo jdbcDatasourceInfo = new JDBCDatasourceInfo();

        // JNDI Name
        String jndiName = getAttribute(objectName, "jndiName");
        jdbcDatasourceInfo.setJndiName(jndiName);

        // URL
        String url = getAttribute(objectName, "url");
        jdbcDatasourceInfo.setUrl(url);

        // JDBC Driver
        String jdbcDriverObjectNameString = getAttribute(objectName, "jdbcDriver");
        ObjectName jdbcDriverObjectName = buildObjectName(jdbcDriverObjectNameString);
        String jdbcDriver = getAttribute(jdbcDriverObjectName, "driverClassName");
        jdbcDatasourceInfo.setJdbcDriver(jdbcDriver);

        // Observable
        boolean isObservable = ((Boolean) getAttribute(objectName, "observable")).booleanValue();
        jdbcDatasourceInfo.setObservable(isObservable);

        // Connections Opened
        int currentConnectionsOpened = ((Integer) getAttribute(objectName, "currentOpened")).intValue();
        jdbcDatasourceInfo.setCurrentConnectionsOpened(currentConnectionsOpened);

        // Current busy
        int currentBusy = ((Integer) getAttribute(objectName, "currentBusy")).intValue();
        jdbcDatasourceInfo.setCurrentBusy(currentBusy);

        // Max pool
        int jdbcMaxConnPool = ((Integer) getAttribute(objectName, "jdbcMaxConnPool")).intValue();
        jdbcDatasourceInfo.setMaxPool(jdbcMaxConnPool);

        return jdbcDatasourceInfo;
    }

    /**
     * @param objectName the JMX objectname of the datasource.
     * @return the datasource information on this datasouce
     */
   public JDBCDatasourceInfo getResourceServiceDatasourceInfo(final ObjectName objectName) {
       JDBCDatasourceInfo jdbcDatasourceInfo = new JDBCDatasourceInfo();

       // JNDI Name
       String jndiName = getAttribute(objectName, "jndiName");
       jdbcDatasourceInfo.setJndiName(jndiName);

       // Get properties of the datasource
       Properties properties = getAttribute(objectName, "properties");

       // URL
       jdbcDatasourceInfo.setUrl(properties.getProperty("URL"));

       // JDBC Driver
       jdbcDatasourceInfo.setJdbcDriver(properties.getProperty("dsClass"));

       // Observable
       boolean isObservable = ((Boolean) getAttribute(objectName, "observable")).booleanValue();
       jdbcDatasourceInfo.setObservable(isObservable);

       // Connections Opened
       int currentConnectionsOpened = ((Integer) getAttribute(objectName, "currentOpened")).intValue();
       jdbcDatasourceInfo.setCurrentConnectionsOpened(currentConnectionsOpened);

       // Current busy
       int currentBusy = ((Integer) getAttribute(objectName, "currentBusy")).intValue();
       jdbcDatasourceInfo.setCurrentBusy(currentBusy);

       // Max pool
       int jdbcMaxConnPool = ((Integer) getAttribute(objectName, "maxSize")).intValue();
       jdbcDatasourceInfo.setMaxPool(jdbcMaxConnPool);

       return jdbcDatasourceInfo;
   }

   /**
    * Gets busy connection infos.
    * @param objectNameString the objectname of the datasource
    * @param duration the duration since getConnection()
    * @return a list of connection with its information
    */
    @SuppressWarnings("unchecked")
    public List<JDBCConnectionInfo> getJDBCConnectionsInfo(final String objectNameString, final int duration) {
        // Get Object
        ObjectName objectName = buildObjectName(objectNameString);
        if (!isRegistered(objectName)) {
            throw new IllegalStateException("No such ObjectName with value '" + objectNameString + "'.");
        }

        // Build list of objects
        List<JDBCConnectionInfo> connectionsInfo = new ArrayList<JDBCConnectionInfo>();

        // Get opened (& filtered) connection IDs
        int[] ids = invoke(objectName, "getOpenedConnections", new Object[] {duration}, new String[] {int.class.getName()});

        // For each connection, get detailed informations
        for (int j = 0; j < ids.length; j++) {
            Map<String, Object> details = invoke(objectName, "getConnectionDetails", new Object[] {ids[j]},
                new String[] {int.class.getName()});

            // Maybe the connection has been closed (and reopen) in the meantime
            Long openDuration = (Long) details.get("duration");
            if ((openDuration == -1 || (openDuration.longValue() >= duration * 1000))) {

                // Add the connection info
                JDBCConnectionInfo jdbcConnectionInfo = new JDBCConnectionInfo();
                connectionsInfo.add(jdbcConnectionInfo);

                jdbcConnectionInfo.setId(((Integer) details.get("id")).intValue());
                Integer openCount = (Integer) details.get("open-count");
                if (openCount != null) {
                    jdbcConnectionInfo.setOpenCount(openCount.intValue());
                } else {
                    // try uses-count
                    Integer usesCount = (Integer) details.get("uses-count");
                    if (usesCount != null) {
                        jdbcConnectionInfo.setOpenCount(usesCount.intValue());
                    }
                }

                Boolean inactive = (Boolean) details.get("inactive");
                if (inactive != null) {
                    jdbcConnectionInfo.setInactive(inactive.booleanValue());
                } else {
                    jdbcConnectionInfo.setInactive(false);
                }
                jdbcConnectionInfo.setDuration(((Long) details.get("duration")).longValue());

                String txId = (String) details.get("transaction-id");
                if (txId != null) {
                    jdbcConnectionInfo.setTransactionXid(txId);
                } else {
                    // try transaction
                    String transaction = (String) details.get("transaction");
                    if (transaction != null) {
                        jdbcConnectionInfo.setTransactionXid(transaction);
                    }
                }
                jdbcConnectionInfo.setAge(((Long) details.get("age")).longValue());

                String txTimeout = (String) details.get("transaction-timeout");
                if (txTimeout != null) {
                    jdbcConnectionInfo.setTransactionTimeout(txTimeout);
                }

                jdbcConnectionInfo.setOpenersThreadInfos(getThreadsInfo((List<Map<String, Object>>) details.get("openers")));
                jdbcConnectionInfo.setClosersThreadInfos(getThreadsInfo((List<Map<String, Object>>) details.get("closers")));

            }
        }
        return connectionsInfo;
    }

    /**
     * Kill the given connection on the given datasource object.
     * @param objectNameString the ObjectName of the datasource
     * @param connectionID the connection's ID
     */
    public void killJDBCConnection(final String objectNameString, final int connectionID) {
        // Get Object
        ObjectName objectName = buildObjectName(objectNameString);
        if (!isRegistered(objectName)) {
            throw new IllegalStateException("No such ObjectName with value '" + objectNameString + "'.");
        }

        // Invoke the close function
        invoke(objectName, "forceCloseConnection", new Object[] {connectionID}, new String[] {int.class.getName()});
    }


    /**
     * @param infos the MBeans data that have been collected
     * @return a list of thread info object
     */
    protected List<ThreadInfo> getThreadsInfo(final List<Map<String, Object>> infos) {
        List<ThreadInfo> threadInfos = new ArrayList<ThreadInfo>();

        for (Iterator<Map<String, Object>> i = infos.iterator(); i.hasNext();) {
            Map<String, Object> info = i.next();

            ThreadInfo threadInfo = new ThreadInfo();
            threadInfos.add(threadInfo);

            // Print general infos about the Thread
            threadInfo.setName((String) info.get("thread.name"));
            threadInfo.setTime(new Date((Long) info.get("thread.time")));
            threadInfo.setStack((String) info.get("thread.stack"));
        }
        return threadInfos;
    }

    /**
     * @return true if the JNDI JDBC Interceptor is registered.
     */
    private boolean isRegistered() {
        return isRegistered(jdbcJNDIInterceptorObjectName);
    }



}
